
extern void print(char *ptr);
extern void itoa(unsigned long n, char *s);

#include "mpc555.h"
#include "tpu_fqd.h"
#include "mpc500_util.h"

inline void tpu_fqd_get_counts(struct TPU3_tag *tpu, UINT8 channel, UINT16 *position, UINT16 *lastedgetime);
void putf (double a, int d);

int main() {

     struct TPU3_tag *tpua = &TPU_A;   /* pointer for TPU routines */
     unsigned long delaycount;
     unsigned short position, lastedgetime, prevposition = 0, prevlastedgetime = 0;
     signed long deltapos, deltatime;
     double velocity;
     unsigned int count;
     char str[50];

    /* Initialize quadrature input function with:   */
    /*    -Input signal on TPU A channel 1 and 2    */
    /*     (note: you just supply the lower
                  (which is primary) channel)       */
    /*    -Inital count value of $0000              */
    /*    -Schedule as high priority in the TPU     */
    tpu_fqd_init( tpua,1 , TPU_PRIORITY_HIGH, 0x0000);
   
    count = 0; 
    while (1) {
	tpu_fqd_get_counts(tpua,1,&position,&lastedgetime);
	if (prevposition > position) { //we wrapped around; assume once
		deltapos = position + (0xffff - prevposition);
		if (deltapos > 30000) {//that was a suspiciously large jump!
			deltapos = position - prevposition;
		}
	}
	else
		deltapos = position - prevposition;

	if (prevlastedgetime > lastedgetime) { //we wrapped around; assume once
		deltatime = lastedgetime + (0xffff - prevlastedgetime);
		//no need to worry b/c time only marches forward! (we hope)
	}
	else
		deltatime = lastedgetime - prevlastedgetime;

	velocity = (1.0*deltapos)/deltatime;

	//deal with non-motion intelligently
	if (deltatime == 0)
		velocity = 0;

	if ((count % 100) == 0) {
	        print("Previous position: ");
        	itoa(prevposition,str);
        	print(str);
        	print(", ");
	        print("Position: ");
        	itoa(position,str);
        	print(str);
        	print("\n");
	        print("Previous last edge timestamp: ");
        	itoa(prevlastedgetime,str);
        	print(str);
        	print(", ");
	        print("Last edge timestamp: ");
        	itoa(lastedgetime,str);
        	print(str);
        	print("\n");
		print("Delta position: ");
//		itoa(deltapos,str);
		putf((double)deltapos, 2);
		print(str);
		print(", ");
		print("Delta time: ");
		itoa(deltatime,str);
		print(str);
		print("\n");
		print("Estimated velocity: ");
		putf(velocity, 3);
		print(" transitions/TPU tick\n");
		print("\n");

//		printf("Previous position: %d, Current position: %d\n", prevposition, position);
//		printf("Previous last edge time: %d, Current last edge time: %d\n", prevlastedgetime, lastedgetime);
//		printf("Estimated velocity: %0.3f transitions/TPU tick\n", velocity);
//		printf("\n");
	}

	prevposition = position;
	prevlastedgetime = lastedgetime;

        delaycount=0;
        do{
                delaycount++;
        }while(delaycount < 5000);
	count++; //wraparound acceptable
    }
}

inline void tpu_fqd_get_counts(struct TPU3_tag *tpu, UINT8 channel, UINT16 *position, UINT16 *lastedgetime)
{
	*position = tpu->PARM.R[channel][TPU_FQD_POSITION_COUNT];
	*lastedgetime = tpu->PARM.R[channel][TPU_FQD_EDGE_TIME];
}

void putf (double a, int d)
{
   int ipart;
   double fpart;
   char tmp[36];

   if (a < 0.0) {
      print("-");
      a *= -1.0;
   }
   /* fpart = modf (a, &ipart); */
   ipart = (int) a;
   fpart = a - (int) a;
   itoa(ipart, tmp);
   print(tmp);
   print(".");
   tmp[1] = '\0';
   for (; d > 0; d--) {
      fpart *= 10.0;
      //fpart = modf (fpart, &ipart);
      ipart = (int) fpart;
      fpart = fpart - (int) fpart;
      tmp[0] = ipart + '0';
      print(tmp);
   }
   itoa((int)fpart, tmp);
   print(tmp);
}

