/*
 * (Hopefully) system-independent implementation of vsprintf & co.
 *
 * Copyright (c) 2004 Michael Dille
 *
 * Derived from code written by others:
 *
 * Copyright (C) 1991, 1992  Linus Torvalds
 * Lars Wirzenius & Linus Torvalds
 * 680x0 support by Hamish Macdonald
 * PPC and vsnprintf suppoort by Kate Alhola
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdarg.h>
#include <ctype.h>

extern void print(char *);


unsigned long simple_strtoul(const char *cp,char **endp,unsigned int base)
{
	unsigned long result = 0,value;

	if (!base) {
		base = 10;
		if (*cp == '0') {
			base = 8;
			cp++;
			if ((*cp == 'x') && isxdigit(cp[1])) {
				cp++;
				base = 16;
			}
		}
	}
	while (isxdigit(*cp) && (value = isdigit(*cp) ? *cp-'0' : (islower(*cp)
	    ? toupper(*cp) : *cp)-'A'+10) < base) {
		result = result*base + value;
		cp++;
	}
	if (endp)
		*endp = (char *)cp;


	return result;
}

/* we use this so that we can do without the ctype library */
#define is_digit(c)	((c) >= '0' && (c) <= '9')

static int skip_atoi(const char **s)
{
	int i=0;

	while (is_digit(**s))
		i = i*10 + *((*s)++) - '0';
	return i;
}

#define ZEROPAD	1		/* pad with zero */
#define SIGN	2		/* unsigned/signed long */
#define PLUS	4		/* show plus */
#define SPACE	8		/* space if plus */
#define LEFT	16		/* left justified */
#define SPECIAL	32		/* 0x */
#define SMALL	64		/* use 'abcdef' instead of 'ABCDEF' */

static char * number(char * str, int num, unsigned int base, int size, int precision
	,int type)
{
	char c,sign,tmp[36];
	const char *digits="0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
	int i;



	if (type&SMALL) digits="0123456789abcdefghijklmnopqrstuvwxyz";
	if (type&LEFT) type &= ~ZEROPAD;
	if (base<2 || base>36)
		return 0;
	c = (type & ZEROPAD) ? '0' : ' ' ;
	if (type&SIGN && num<0) {
		sign='-';
		num = -num;
	} else
		sign=(type&PLUS) ? '+' : ((type&SPACE) ? ' ' : 0);
	if (sign) size--;


	if (type&SPECIAL)
		if (base==16) size -= 2;
		else if (base==8) size--;
	i=0;

	if (num==0)
		tmp[i++]='0';
	else while (num!=0) {
	  tmp[i++]=digits[(unsigned int)num % base];
	  (unsigned int)num=(unsigned int)num/base;
	};


	if (i>precision) precision=i;
	size -= precision;
	if (!(type&(ZEROPAD+LEFT)))
		while(size-->0)
			*str++ = ' ';
	if (sign)
		*str++ = sign;
	if (type&SPECIAL)
		if (base==8)
			*str++ = '0';
		else if (base==16) {
			*str++ = '0';
			*str++ = digits[33];
		}
	if (!(type&LEFT))
		while(size-->0)
			*str++ = c;
	while(i<precision--)
		*str++ = '0';
	while(i-->0)
		*str++ = tmp[i];
	while(size-->0)
		*str++ = ' ';


	return str;
}


/* (At least temporarily) has the limitation that the integer portion of the
   floating point number must be representable as an int.  This is because I
   sought to avoid any use of libm functions. */
/* Ignores SPECIAL flag */
static char * fnumber(char * str, double num, unsigned int base, int size, int precision
	,int type)
{
	char sign,c,tmpi[36], tmpf[36];
	int ipart;
	double fpart;
	int i, j;
	const char *digits="0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
	int carry;
	
	if (type&SMALL) digits="0123456789abcdefghijklmnopqrstuvwxyz";

	if (type&LEFT) type &= ~ZEROPAD;
	c = (type & ZEROPAD) ? '0' : ' ' ;
	if (type&SIGN && num<0.0) {
		sign='-';
		num = -1.0*num;
	} else
		sign=(type&PLUS) ? '+' : ((type&SPACE) ? ' ' : 0);
	if (sign) size--;

	fpart = num - (int) num;

	for (j = 0; j < precision; j++) {
		fpart *= 1.0*base;
		ipart = (int) fpart;
		fpart = fpart - (int) fpart;
		tmpf[j]=digits[(unsigned int)ipart % base];
	}

	ipart = (int) num;

	carry = (fpart >= 0.5);
	i = j-1;
	while (carry > 0) {
		if (i<0) {
			ipart++; /* must increment integer component */
			break;
		}
		tmpf[i]++;
		carry = (tmpf[i]-digits[0])/base;
		tmpf[i] = digits[(tmpf[i]-digits[0])%base];
		i--;
	}

	i = 0;
	if (ipart==0)
		tmpi[i++]='0';
	else while (ipart!=0) {
	  tmpi[i++]=digits[(unsigned int)ipart % base];
	  (unsigned int)ipart=(unsigned int)ipart/base;
	};


	size -= i+j+(precision > 0); /* don't forget decimal point! */

	if (!(type&(ZEROPAD+LEFT)))
		while(size-->0)
			*str++ = ' ';
	if (sign)
		*str++ = sign;
	if (!(type&LEFT))
		while(size-->0)
			*str++ = c;

	/* copy integer portion */
	while(i-->0)
		*str++ = tmpi[i];

	if (precision > 0)
		*str++ = '.';
	
	/* copy float portion; note it is not reversed */
	for (j = 0; j < precision; j++)
		*str++ = tmpf[j];

	while(size-->0)
		*str++ = ' ';

	return str;
}


int vsnprintf(char *buf, int size, const char *fmt, va_list args)
{
	int len;
	int i;
	char * str;
	char *s;
	int *ip;

	int flags;		/* flags to number() */

	int field_width;	/* width of output field */
	int precision;		/* min. # of digits for integers; max
				   number of chars for from string */
	int qualifier;		/* 'h', 'l', or 'L' for integer fields */



	for (str=buf ; *fmt &&  ((str-buf)<size) ; ++fmt) {
		if (*fmt != '%') {
			*str++ = *fmt;
			continue;
		}
			
		/* process flags */
		flags = 0;
		repeat:
			++fmt;		/* this also skips first '%' */
			switch (*fmt) {
				case '-': flags |= LEFT; goto repeat;
				case '+': flags |= PLUS; goto repeat;
				case ' ': flags |= SPACE; goto repeat;
				case '#': flags |= SPECIAL; goto repeat;
				case '0': flags |= ZEROPAD; goto repeat;
				}
		
		/* get field width */
		field_width = -1;
		if (is_digit(*fmt))
			field_width = skip_atoi(&fmt);
		else if (*fmt == '*') {
			/* it's the next argument */
			field_width = va_arg(args, int);
			if (field_width < 0) {
				field_width = -field_width;
				flags |= LEFT;
			}
			if(field_width>(str-buf)) field_width=str-buf;
		}

		/* get the precision */
		precision = -1;
		if (*fmt == '.') {
			++fmt;	
			if (is_digit(*fmt))
				precision = skip_atoi(&fmt);
			else if (*fmt == '*') {
				/* it's the next argument */
				precision = va_arg(args, int);
			}
			if (precision < 0)
				precision = 0;
		}

		/* get the conversion qualifier */
		qualifier = -1;
		if (*fmt == 'h' || *fmt == 'l' || *fmt == 'L') {
			qualifier = *fmt;
			++fmt;
		}

		switch (*fmt) {
		case 'c':
			if (!(flags & LEFT))
				while (--field_width > 0)
					*str++ = ' ';
			*str++ = (unsigned char) va_arg(args, int);
			while (--field_width > 0)
				*str++ = ' ';
			break;

		case 's':
			s = va_arg(args, char *);
			if (!s)
				s = "<NULL>";
			len = strlen(s);
			if (precision < 0)
				precision = len;
			else if (len > precision)
				len = precision;

			if (!(flags & LEFT))
			     while ((len < field_width--) && ((str-buf)<size))
					*str++ = ' ';
			for (i = 0; (i < len) && ((str-buf)<size); ++i)
				*str++ = *s++;
			while ((len < field_width--)  &&  ((str-buf)<size))
				*str++ = ' ';
			break;


		case 'o':
			str = number(str, va_arg(args, unsigned long), 8,
				field_width, precision, flags);
				break; 

		case 'p':
			if (field_width == -1) {
				field_width = 8;
				flags |= ZEROPAD;
			}
			str = number(str,
				(unsigned long) va_arg(args, void *), 16,
				field_width, precision, flags);
			break;

		case 'x':
			flags |= SMALL;
		case 'X':
			str = number(str, va_arg(args, unsigned long), 16,
				field_width, precision, flags);
			break;

		case 'd':
		case 'i':
			flags |= SIGN;
		case 'u':
			str = number(str, va_arg(args, unsigned long), 10,
				field_width, precision, flags);
			break;

		case 'n':
			ip = va_arg(args, int *);
			*ip = (str - buf);
			break;
		case 'f':
			flags |= SIGN;
			if (precision < 0) precision = 6; /* default */
			str = fnumber(str, va_arg(args, double), 10,
				field_width, precision, flags);
			break;
		default:
			if (*fmt != '%')
				*str++ = '%';
			if (*fmt)
				*str++ = *fmt;
			else
				--fmt;
			break;
		}
	}
	*str = '\0';
	return str-buf;
}

int sprintf(char * buf, const char *fmt, ...)
{
	va_list args;
	int i;


	va_start(args, fmt);
	i=vsnprintf(buf,256,fmt,args);
	va_end(args);
	return i;
}

int snprintf(char * buf,int size,const char *fmt, ...)
{
	va_list args;
	int i;


	va_start(args, fmt);
	i=vsnprintf(buf,size,fmt,args);
	va_end(args);
	return i;
}
/*======================================================
!
! printf to console using debug printf routine
!
\*======================================================*/


int printf(const char *fmt, ...)
{
	va_list args;
	int i;
	static char buf[120];

	va_start(args, fmt);
	i=vsnprintf(buf,sizeof(buf),fmt,args);
	print(buf); //use low-level libc print
	va_end(args);
	return i;
}




